#! /usr/bin/env python

from setuptools import setup

setup(
    name='sco-client',
    version='0.1.0',
    description='Library to interact with existing Standard Cortical Observer Web API',
    keywords='neuroscience vision cortex ',
    author='Heiko Mueller',
    author_email='heiko.muller@gmail.com',
    url='https://github.com/heikomuller/sco-client',
    license='GPLv3',
    packages=['scocli'],
    package_data={'': ['LICENSE']},
    install_requires=['python-dateutil',
                      'requests']
)



"""Collection of methods and definitions related to the Standard Cortical
Observer Web API resources.
"""

import datetime as dt
from dateutil import tz
import json
import os
import requests
import shutil
import tarfile
import tempfile
import urllib2


# ------------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------------

"""HATEOAS reference keys."""

# SCO-API create experiment
REF_EXPERIMENTS_CREATE = 'experiments.create'
# SCO-API experiments listing
REF_EXPERIMENTS_LISTING = 'experiments.list'
# SCO-API list experiments runs
REF_EXPERIMENTS_RUNS_CREATE = 'predictions.run'
# SCO-API list experiments runs
REF_EXPERIMENTS_RUNS_LISTING = 'predictions.list'
# SCO-API create image group
REF_IMAGE_GROUPS_CREATE = 'images.upload'
# SCO-API image groups listing
REF_IMAGE_GROUPS_LIST = 'images.groups.list'
# SCO-API create subject
REF_SUBJECTS_CREATE = 'subjects.upload'
# SCO-API subjects listing
REF_SUBJECTS_LIST = 'subjects.list'
# Resource download
REF_DOWNLOAD = 'download'
# Resource links listing
REF_LINKS = 'links'
# Resource self reference
REF_SELF = 'self'
# Upsert options (currently for image groups only)
REF_UPDATE_OPTIONS = 'options'
# Update model run state
REF_UPDATE_STATE = 'state'
# Upsert properties reference for resources
REF_UPSERT_PROPERTIES = 'properties'

"""Query parameter for object listings."""

# List of attributes to include for each item in listings
QPARA_ATTRIBUTES = 'properties'
# Limit number of items in result
QPARA_LIMIT = 'limit'
# Set offset in collection
QPARA_OFFSET = 'offset'

"""Model run timestamp keys"""

RUN_CREATED_AT = 'createdAt'
RUN_FINISHED_AT = 'finishedAt'
RUN_STARTED_AT = 'startedAt'

""" Run states """

RUN_FAILED = 'FAILED'
RUN_IDLE = 'IDLE'
RUN_ACTIVE = 'RUNNING'
RUN_SUCCESS = 'SUCCESS'

# ------------------------------------------------------------------------------
#
# SCO Web API Resources
#
# ------------------------------------------------------------------------------

class ResourceHandle(object):
    """Generic handle for a Web API resource in resource listing. Contains the
    four basic resource attributes identifier, name, timestamp, and url. If
    additional properties where requested in the listing call, these will be
    available in a properties dictionary.

    Attributes
    ----------
    identifier : string
        Unique resource identifier
    name : string
        Resource name
    timestamp : datetime.datetime
        Timestamp of resource creation (UTC)
    url : string
        Url to access the resource
    properties : Dictionary
        Dicrionary of additional resource properties and their values. None if
        no additional properties where requested in the client listing method
        call.
    links : Dictionary
        Dictionary of HATEOAS references associated with the resource
    """
    def __init__(self, json_obj):
        """Initialize the resource handle using the Json object for the resource
        in the listing result returned by the Web API.

        Parameters
        ----------
        json_obj : Json object
            Json object for resources as returned by Web API
        """
        # Get resource attributes from the Json object
        self.identifier = json_obj['id']
        self.name = json_obj['name']
        # Convert object's creation timestamp from UTC to local time
        self.timestamp = to_local_time(json_obj['timestamp'])
        # Get resource HATEOAS references
        self.links = references_to_dict(json_obj[REF_LINKS])
        # Get self reference from list of resource links
        self.url = self.links[REF_SELF]
        # Set resource properties if present in the Json object. For handles
        # in object listings the property element will not be present. In that
        # case the local attribute is set to None.
        if 'properties' in json_obj:
            self.properties = {}
            for kvp in json_obj['properties']:
                self.properties[str(kvp['key'])] = str(kvp['value'])
        else:
            self.properties = None


class ExperimentHandle(ResourceHandle):
    """Resource handle for SCO experiment resource. Experiments are not directly
    associated with any downloadable data files. However, the experiment refers
    to associated subject and image group resources that are cached on local
    disk.

    Attributes
    ----------
    fmri_url : string
        Url for associated fMRI resource. None if no fMRI is associated with
        this experiment
    image_group_url : string
        Url for associated image group resource.
    runs_url : string
        Url for associated model runs
    subject_url : string
        Url for associated subject resource
    """
    def __init__(self, json_obj, sco):
        """Initialize image group handle.
        Parameters
        ----------
        json_obj : Json-like object
            Json object containing resource description
        sco : SCOClient
            Client to access associated resources.
        """
        super(ExperimentHandle, self).__init__(json_obj)
        # Maintain reference to SCO client to access subject and image group
        # resources when requested
        self.sco = sco
        # Maintain Urls for associated subject, image group, fMRI, and model
        # run resources
        self.subject_url = references_to_dict(
            json_obj['subject']['links']
        )[REF_SELF]
        self.image_group_url = references_to_dict(
            json_obj['images']['links']
        )[REF_SELF]
        if 'fmri' in json_obj:
            self.fmri_url = references_to_dict(
                json_obj['fmri']['links']
            )[REF_SELF]
        else:
            self.fmri_url = None
        self.runs_url = self.links[REF_EXPERIMENTS_RUNS_LISTING]

    @staticmethod
    def create(url, name, subject_id, image_group_id, properties):
        """Create a new experiment using the given SCO-API create experiment Url.

        Parameters
        ----------
        url : string
            Url to POST experiment create request
        name : string
            User-defined name for experiment
        subject_id : string
            Unique identifier for subject at given SCO-API
        image_group_id : string
            Unique identifier for image group at given SCO-API
        properties : Dictionary
            Set of additional properties for created experiment. Argument may be
            None. Given name will override name property in this set (if present).

        Returns
        -------
        string
            Url of created experiment resource
        """
        # Create list of key,value-pairs representing experiment properties for
        # request. The given name overrides the name in properties (if present).
        obj_props = [{'key':'name','value':name}]
        if not properties is None:
            # Catch TypeErrors if properties is not a list.
            try:
                for key in properties:
                    if key != 'name':
                        obj_props.append({'key':key, 'value':properties[key]})
            except TypeError as ex:
                raise ValueError('invalid property set')
        # Create request body and send POST request to given Url
        body = {
            'subject' : subject_id,
            'images' : image_group_id,
            'properties' : obj_props
        }
        try:
            req = urllib2.Request(url)
            req.add_header('Content-Type', 'application/json')
            response = urllib2.urlopen(req, json.dumps(body))
        except urllib2.URLError as ex:
            raise ValueError(str(ex))
        # Get experiment self reference from successful response
        return references_to_dict(json.load(response)['links'])[REF_SELF]

    @property
    def image_group(self):
        """Image group resource that is associated with this experiment.

        Returns
        -------
        ImageGroupHandle
            Handle for associated image group resource
        """
        return self.sco.image_groups_get(self.image_group_url)

    def run(self, name, arguments={}, properties=None):
        """Create a new model run with given name, arguments, and properties.
        Parameters
        ----------
        name : string
            User-defined name for experiment
        arguments : Dictionary
            Dictionary of arguments for model run
        properties : Dictionary, optional
            Set of additional properties for created mode run.

        Returns
        -------
        scoserv.ModelRunHandle
            Handle for local copy of created model run resource
        """
        return self.sco.experiments_runs_create(
            name,
            self.links[REF_EXPERIMENTS_RUNS_CREATE],
            arguments=arguments,
            properties=properties
        )

    def runs(self, offset=0, limit=-1, properties=None):
        """Get a list of run descriptors associated with this expriment.

        Parameters
        ----------
        offset : int, optional
            Starting offset for returned list items
        limit : int, optional
            Limit the number of items in the result
        properties : List(string)
            List of additional object properties to be included for items in
            the result

        Returns
        -------
        List(scoserv.ModelRunDescriptor)
            List of model run descriptors
        """
        return get_run_listing(
            self.runs_url,
            offset=offset,
            limit=limit,
            properties=properties
        )

    @property
    def subject(self):
        """Subject resource that is associated with this experiment.

        Returns
        -------
        SubjectHandle
            Handle for associated subject resource
        """
        return self.sco.subjects_get(self.subject_url)


class ImageGroupHandle(ResourceHandle):
    """Resource handle for SCO image group resource on local disk. The contents
    of the image group tar-file are extracted into the objects data directory.

    Attributes
    ----------
    data_dir : string
        Absolute path to directory containing a local copy of the resource
        data files
    images : List(string)
        List if absolute file path to images in the group.
    options : Dictionary
        Dictionary of options for image group
    """
    def __init__(self, json_obj, base_dir):
        """Initialize image group handle.
        Parameters
        ----------
        json_obj : Json-like object
            Json object containing resource description
        base_dir : string
            Path to cache base directory for object
        """
        super(ImageGroupHandle, self).__init__(json_obj)
        # Set image group options
        self.options = {}
        for kvp in json_obj['options']:
            self.options[str(kvp['name'])] = kvp['value']
        # Set the data directory. If directory does not exist, create it,
        # download the resource data archive, and unpack into data directory
        self.data_dir = os.path.abspath(os.path.join(base_dir, 'data'))
        if not os.path.isdir(self.data_dir):
            os.mkdir(self.data_dir)
            # Download tar-archive
            tmp_file, f_suffix = download_file(self.links[REF_DOWNLOAD])
            # Unpack downloaded file into data directory
            try:
                tf = tarfile.open(name=tmp_file, mode='r')
                tf.extractall(path=self.data_dir)
            except (tarfile.ReadError, IOError) as err:
                # Clean up in case there is an error during extraction
                shutil.rmtree(self.data_dir)
                raise ValueError(str(err))
            # Remove downloaded file
            os.remove(tmp_file)
        # Set list of group images. The list (and order) of images is stored in
        # the .images file in the resource's data directory. If the file does
        # not exists read list from SCO-API.
        self.images = []
        images_file = os.path.join(base_dir, '.images')
        if not os.path.isfile(images_file):
            json_list = JsonResource(
                references_to_dict(
                    json_obj['images']['links']
                )[REF_SELF] + '?' + QPARA_LIMIT + '=-1'
            ).json
            with open(images_file, 'w') as f:
                for element in json_list['items']:
                    # Folder names start with '/'. Remove to get abs local path
                    local_path = element['folder'][1:] + element['name']
                    img_file = os.path.join(self.data_dir, local_path)
                    f.write(local_path + '\n')
                    self.images.append(img_file)
        else:
            # Read content of images file into images list
            with open(images_file, 'r') as f:
                for line in f:
                    self.images.append(os.path.join(self.data_dir, line.strip()))

    @staticmethod
    def create(url, filename, options, properties):
        """Create new image group at given SCO-API by uploading local file.
        Expects an tar-archive containing images in the image group. Allows to
        update properties of created resource.

        Parameters
        ----------
        url : string
            Url to POST image group create request
        filename : string
            Path to tar-archive on local disk
        options : Dictionary, optional
            Values for image group options. Argument may be None.
        properties : Dictionary
            Set of additional properties for image group (may be None)

        Returns
        -------
        string
            Url of created image group resource
        """
        # Ensure that the file has valid suffix
        if not has_tar_suffix(filename):
            raise ValueError('invalid file suffix: ' + filename)
        # Upload file to create image group. If response is not 201 the uploaded
        # file is not a valid tar file
        files = {'file': open(filename, 'rb')}
        response = requests.post(url, files=files)
        if response.status_code != 201:
            raise ValueError('invalid file: ' + filename)
        # Get image group HATEOAS references from successful response
        links = references_to_dict(response.json()['links'])
        resource_url = links[REF_SELF]
        # Update image group options if given
        if not options is None:
            obj_ops = []
            # Catch TypeErrors if properties is not a list.
            try:
                for opt in options:
                    obj_ops.append({'name' : opt, 'value' : options[opt]})
            except TypeError as ex:
                raise ValueError('invalid option set')
            try:
                req = urllib2.Request(links[REF_UPDATE_OPTIONS])
                req.add_header('Content-Type', 'application/json')
                response = urllib2.urlopen(
                    req,
                    json.dumps({'options' : obj_ops})
                )
            except urllib2.URLError as ex:
                raise ValueError(str(ex))
        # Update image group properties if given
        if not properties is None:
            obj_props = []
            # Catch TypeErrors if properties is not a list.
            try:
                for key in properties:
                    obj_props.append({'key':key, 'value':properties[key]})
            except TypeError as ex:
                raise ValueError('invalid property set')
            try:
                req = urllib2.Request(links[REF_UPSERT_PROPERTIES])
                req.add_header('Content-Type', 'application/json')
                response = urllib2.urlopen(
                    req,
                    json.dumps({'properties' : obj_props})
                )
            except urllib2.URLError as ex:
                raise ValueError(str(ex))
        return resource_url


class ModelRunDescriptor(ResourceHandle):
    """Handle for model runs. Extends the default resource handle with
    information about the run state.

    Attributes
    ----------
    state : string
        State of the model run ('FAILED', 'IDLE', 'RUNNING', 'SUCCESS')
    """
    def __init__(self, json_obj):
        """Initialize the run descriptor using the Json object for the model
        run in the listing result returned by the Web API.

        Parameters
        ----------
        json_obj : Json object
            Json object for model run as returned by Web API. Expected to
            contain additional state field.
        """
        super(ModelRunDescriptor, self).__init__(json_obj)
        # Set run state
        self.state = json_obj['state']


class ModelRunHandle(ModelRunDescriptor):
    """Resource handle for SCO model run. Contains dictionary of timestamps
    for run scheduling. For failed runs a list of error messages is maintained.
    For completed runs a rference to the result file is maintainted.

    Attributes
    ----------
    arguments : Dictionary()
        Dictionary of arguments for the model run
    experiment_url : string
        Url for associated experiment
    schedule : Dictionary(datetime)
        Dictionary of timestamps for run events
    errors : List(string), Optional
        List of error messages (only for runs in FAILED state)
    result_file : string
        Path to result file (only for runs in SUCCESS state)
    """
    def __init__(self, json_obj, base_dir, sco):
        """Initialize subject handle.
        Parameters
        ----------
        json_obj : Json-like object
            Json object containing resource description
        base_dir : string
            Path to cache base directory for object
        sco : SCOClient
            Client to access associated experiment.
        """
        super(ModelRunHandle, self).__init__(json_obj)
        # Maintain reference to SCO client to access experiment
        self.sco = sco
        # Create list of arguments for model run
        # Set image group options
        self.arguments = {}
        for kvp in json_obj['arguments']:
            self.arguments[str(kvp['name'])] = kvp['value']
        # Extract experiment Url
        self.experiment_url = references_to_dict(
            json_obj['experiment'][REF_LINKS]
        )[REF_SELF]
        # Create list of schedule timestamps
        self.schedule = {}
        for key in json_obj['schedule']:
            self.schedule[key] = to_local_time(json_obj['schedule'][key])
        # For failed runs create attribute errors containing list of errors
        # messages
        if self.state == RUN_FAILED:
            self.errors = json_obj['errors']
        elif self.state == RUN_SUCCESS:
            # Name of the result data file
            filename = None
            # Create data directory if it doesnt exist
            data_dir = os.path.abspath(os.path.join(base_dir, 'data'))
            if not os.path.isdir(data_dir):
                os.mkdir(data_dir)
            else:
                # If the directory exists check whether it contains the result
                # data file. The file name is expected to start with the run
                # identifier.
                for f in os.listdir(data_dir):
                    if f.startswith(self.identifier):
                        filename = f
                        break
            # Download the result file if it has not been downloaded yet
            if filename is None:
                tmp_file, f_suffix = download_file(self.links[REF_DOWNLOAD])
                filename = self.identifier + f_suffix
                self.result_file = os.path.join(data_dir, filename)
                shutil.move(tmp_file, self.result_file)
            else:
                self.result_file = os.path.join(data_dir, filename)

    @staticmethod
    def create(url, name, arguments, properties=None):
        """Create a new model run using the given SCO-API create model run Url.

        Parameters
        ----------
        url : string
            Url to POST model run create model run request
        name : string
            User-defined name for model run
        arguments : Dictionary
            Dictionary of arguments for model run
        properties : Dictionary, optional
            Set of additional properties for created mode run.

        Returns
        -------
        string
            Url of created model run resource
        """
        # Create list of model run arguments. Catch TypeErrors if arguments is
        # not a list.
        obj_args = []
        try:
            for arg in arguments:
                obj_args.append({'name' : arg, 'value' : arguments[arg]})
        except TypeError as ex:
            raise ValueError('invalid argument set')
        # Create request body and send POST request to given Url
        body = {
            'name' : name,
            'arguments' : obj_args,
        }
        # Create list of properties if given.  Catch TypeErrors if properties is
        # not a list.
        if not properties is None:
            obj_props = []
            try:
                for key in properties:
                    if key != 'name':
                        obj_props.append({'key':key, 'value':properties[key]})
            except TypeError as ex:
                raise ValueError('invalid property set')
            body['properties'] =  obj_props
        # POST create model run request
        try:
            req = urllib2.Request(url)
            req.add_header('Content-Type', 'application/json')
            response = urllib2.urlopen(req, json.dumps(body))
        except urllib2.URLError as ex:
            raise ValueError(str(ex))
        # Get model run self reference from successful response
        return references_to_dict(json.load(response)['links'])[REF_SELF]

    @property
    def experiment(self):
        """Experiment resource for which this is a model run.

        Returns
        -------
        ExperimentHandle
            Handle for associated experiment resource
        """
        return self.sco.experiments_get(self.experiment_url)

    def refresh(self):
        """Get a refreshed version of the resource handle. Primarily necessary
        to minitor changes to the run state.

        Note that this handle is not refreshed but a fresh handle is returned!

        Returns
        -------
        ModelRunHandle
            Refreshed run handle.
        """
        return self.sco.experiments_runs_get(self.url)

    @staticmethod
    def update_state(url, state_obj):
        """Update the state of a given model run. The state object is a Json
        representation of the state as created by the SCO-Server.

        Throws a ValueError if the resource is unknown or the update state
        request failed.

        Parameters
        ----------
        url : string
            Url to POST model run create model run request
        state_obj : Json object
            State object serialization as expected by the API.
        """
        # POST update run state request
        try:
            req = urllib2.Request(url)
            req.add_header('Content-Type', 'application/json')
            response = urllib2.urlopen(req, json.dumps(state_obj))
        except urllib2.URLError as ex:
            raise ValueError(str(ex))
        # Throw exception if resource was unknown or update request failed
        if response.code == 400:
            raise ValueError(response.message)
        elif response.code == 404:
            raise ValueError('unknown model run')

    def update_state_active(self):
        """Update the state of the model run to active.

        Raises an exception if update fails or resource is unknown.

        Returns
        -------
        ModelRunHandle
            Refreshed run handle.
        """
        # Update state to active
        self.update_state(self.links[REF_UPDATE_STATE], {'type' : RUN_ACTIVE})
        # Returned refreshed verion of the handle
        return self.refresh()

    def update_state_error(self, errors):
        """Update the state of the model run to 'FAILED'. Expects a list of
        error messages.

        Raises an exception if update fails or resource is unknown.

        Parameters
        ----------
        errors : List(string)
            List of error messages

        Returns
        -------
        ModelRunHandle
            Refreshed run handle.
        """
        # Update state to active
        self.update_state(
            self.links[REF_UPDATE_STATE],
            {'type' : RUN_FAILED, 'errors' : errors}
        )
        # Returned refreshed verion of the handle
        return self.refresh()

    def update_state_success(self, model_output):
        """Update the state of the model run to 'SUCCESS'. Expects a model
        result identifier.

        Raises an exception if update fails or resource is unknown.

        Parameters
        ----------
        model_output : string
            Identifier of fMRI data object

        Returns
        -------
        ModelRunHandle
            Refreshed run handle.
        """
        # Update state to active
        self.update_state(
            self.links[REF_UPDATE_STATE],
            {'type' : RUN_SUCCESS, 'modelOutput' : model_output}
        )
        # Returned refreshed verion of the handle
        return self.refresh()


class SubjectHandle(ResourceHandle):
    """Resource handle for SCO subject resource on local disk. Downloads the
    subject tar-file (on first access) and copies the contained FreeSurfer
    directory into the resource's data directory.


    Attributes
    ----------
    data_dir : string
        Absolute path to directory containing the subjects FreeSurfer data files
    """
    def __init__(self, json_obj, base_dir):
        """Initialize subject handle.
        Parameters
        ----------
        json_obj : Json-like object
            Json object containing resource description
        base_dir : string
            Path to cache base directory for object
        """
        super(SubjectHandle, self).__init__(json_obj)
        # Set the data directory. If directory does not exist, create it,
        # download the resource data archive, and unpack into data directory
        self.data_dir = os.path.abspath(os.path.join(base_dir, 'data'))
        if not os.path.isdir(self.data_dir):
            # Create data dir and temporary directory to extract downloaded file
            os.mkdir(self.data_dir)
            temp_dir = tempfile.mkdtemp()
            # Download tar-archive and unpack into temp_dir
            tmp_file, f_suffix = download_file(self.links[REF_DOWNLOAD])
            try:
                tf = tarfile.open(name=tmp_file, mode='r')
                tf.extractall(path=temp_dir)
            except (tarfile.ReadError, IOError) as err:
                # Clean up in case there is an error during extraction
                shutil.rmtree(temp_dir)
                shutil.rmtree(self.data_dir)
                raise ValueError(str(err))
            # Remove downloaded file
            os.remove(tmp_file)
            # Make sure the extracted files contain a valid freesurfer directory
            freesurf_dir = get_freesurfer_dir(temp_dir)
            if not freesurf_dir:
                # Clean up before raising an exception
                shutil.rmtree(temp_dir)
                shutil.rmtree(self.data_dir)
                raise ValueError('not a valid subject directory')
            # Move all sub-folders from the Freesurfer directory to the new anatomy
            # data directory
            for f in os.listdir(freesurf_dir):
                sub_folder = os.path.join(freesurf_dir, f)
                if os.path.isdir(sub_folder):
                    shutil.move(sub_folder, self.data_dir)
            # Remove temporary directory
            shutil.rmtree(temp_dir)


    @staticmethod
    def create(url, filename, properties):
        """Create new subject at given SCO-API by uploading local file.
        Expects an tar-archive containing FreeSurfer archive file. Allows to
        update properties of created resource.

        Parameters
        ----------
        url : string
            Url to POST image group create request
        filename : string
            Path to tar-archive on local disk
        properties : Dictionary
            Set of additional properties for subject (may be None)

        Returns
        -------
        string
            Url of created subject resource
        """
        # Ensure that the file has valid suffix
        if not has_tar_suffix(filename):
            raise ValueError('invalid file suffix: ' + filename)
        # Upload file to create subject. If response is not 201 the uploaded
        # file is not a valid FreeSurfer archive
        files = {'file': open(filename, 'rb')}
        response = requests.post(url, files=files)
        if response.status_code != 201:
            raise ValueError('invalid file: ' + filename)
        # Get image group HATEOAS references from successful response
        links = references_to_dict(response.json()['links'])
        resource_url = links[REF_SELF]
        # Update subject properties if given
        if not properties is None:
            obj_props = []
            # Catch TypeErrors if properties is not a list.
            try:
                for key in properties:
                    obj_props.append({'key':key, 'value':properties[key]})
            except TypeError as ex:
                raise ValueError('invalid property set')
            try:
                req = urllib2.Request(links[REF_UPSERT_PROPERTIES])
                req.add_header('Content-Type', 'application/json')
                response = urllib2.urlopen(
                    req,
                    json.dumps({'properties' : obj_props})
                )
            except urllib2.URLError as ex:
                raise ValueError(str(ex))
        return resource_url


# ------------------------------------------------------------------------------
#
# Helper Classes
#
# ------------------------------------------------------------------------------

class JsonResource:
    """Simple class to wrap a GET request that reads a Json object. Includes the
    request response and the retrieved Json object.

    Attributes
    ----------
    json : Json object
        Json response object
    response : Response
        Http request response object
    """
    def __init__(self, url):
        """Get Json object from given Url.

        Raises ValueError if given Url cannot be read or result is not a valid
        Json object.

        Parameters
        ----------
        url : string
            Url of resource to be read
        """
        try:
            self.response = urllib2.urlopen(url)
        except urllib2.URLError as ex:
            raise ValueError(str(ex))
        self.json = json.loads(self.response.read())


# ------------------------------------------------------------------------------
#
# Helper Methods
#
# ------------------------------------------------------------------------------

def download_file(url):
    """Download attached file as temporary file.

    Parameters
    ----------
    url : string
        SCO-API downlioad Url

    Returns
    -------
    string, string
        Path to downloaded file and file suffix
    """
    r = urllib2.urlopen(url)
    # Expects a tar-archive or a compressed tar-archive
    if r.info()['content-type'] == 'application/x-tar':
        suffix = '.tar'
    elif r.info()['content-type'] == 'application/gzip':
        suffix = '.tar.gz'
    else:
        raise ValueError('unexpected file type: ' + r.info()['content-type'])
    # Save attached file in temp file and return path to temp file
    fd, f_path = tempfile.mkstemp(suffix=suffix)
    os.write(fd, r.read())
    os.close(fd)
    return f_path, suffix


def get_freesurfer_dir(directory):
    """Test if a directory is a Freesurfer anatomy directory. Currently, the
    test is whether (1) there are sub-folders with name 'surf' and 'mri' and
    (2) if the Freesurfer library method freesurfer_subject returns a non-None
    result for the directory. Processes all sub-folders recursively until a
    freesurfer directory is found. If no matching folder is found the result is
    None.

    Parameters
    ----------
    directory : string
        Directory on local disk containing unpacked files

    Returns
    -------
    string
        Sub-directory containing a Freesurfer files or None if no such
        directory is found.
    """
    dir_files = [f for f in os.listdir(directory)]
    # Look for sub-folders 'surf' and 'mri'
    if 'surf' in dir_files and 'mri' in dir_files:
        return directory
    # Directory is not a valid freesurfer directory. Continue to search
    # recursively until a matching directory is found.
    for f in os.listdir(directory):
        sub_dir = os.path.join(directory, f)
        if os.path.isdir(sub_dir):
            if get_freesurfer_dir(sub_dir):
                return sub_dir
    # The given directory does not contain a freesurfer anatomy directory
    return None


def get_resource_listing(url, offset, limit, properties):
    """Gneric method to retrieve a resource listing from a SCO-API. Takes the
    resource-specific API listing Url as argument.

    Parameters
    ----------
    url : string
        Resource listing Url for a SCO-API
    offset : int, optional
        Starting offset for returned list items
    limit : int, optional
        Limit the number of items in the result
    properties : List(string)
        List of additional object properties to be included for items in
        the result

    Returns
    -------
    List(ResourceHandle)
        List of resource handle (one per subject in the object listing)
    """
    # Create listing query based on given arguments
    query = [
        QPARA_OFFSET + '=' + str(offset),
        QPARA_LIMIT + '=' + str(limit)
    ]
    # Add properties argument if property list is not None and not empty
    if not properties is None:
        if len(properties) > 0:
            query.append(QPARA_ATTRIBUTES + '=' + ','.join(properties))
    # Add query to Url.
    url = url + '?' + '&'.join(query)
    # Get subject listing Url for given SCO-API and decorate it with
    # given listing arguments. Then retrieve listing from SCO-API.
    json_obj = JsonResource(url).json
    # Convert result into a list of resource handles and return the result
    resources = []
    for element in json_obj['items']:
        resource = ResourceHandle(element)
        # Add additional properties to resource if list is given
        if not properties is None:
            resource.properties = {}
            for prop in properties:
                if prop in element:
                    resource.properties[prop] = element[prop]
        resources.append(resource)
    return resources


def get_run_listing(listing_url, offset, limit, properties):
    """Get list of experiment resources from a SCO-API.

    Parameters
    ----------
    listing_url : string
        url for experiments run listing.
    offset : int
        Starting offset for returned list items
    limit : int
        Limit the number of items in the result
    properties : List(string)
        List of additional object properties to be included for items in
        the result

    Returns
    -------
    List(scoserv.ModelRunDescriptor)
        List of model run descriptors
    """
    # Create listing query based on given arguments
    query = [
        QPARA_OFFSET + '=' + str(offset),
        QPARA_LIMIT + '=' + str(limit)
    ]
    # Ensure that the run state is included in the listing as attribute
    props = ['state']
    # Add properties argument if property list is not None and not empty
    if not properties is None:
        for prop in properties:
            if not prop in props:
                props.append(prop)
    query.append(QPARA_ATTRIBUTES + '=' + ','.join(props))
    # Add query to Url.
    url = listing_url + '?' + '&'.join(query)
    # Get subject listing Url for given SCO-API and decorate it with
    # given listing arguments. Then retrieve listing from SCO-API.
    json_obj = JsonResource(url).json
    # Convert result into a list of resource handles and return the result
    resources = []
    for element in json_obj['items']:
        resource = ModelRunDescriptor(element)
        # Add additional properties to resource if list is given
        if not properties is None:
            resource.properties = {}
            for prop in properties:
                if prop in element:
                    resource.properties[prop] = element[prop]
        resources.append(resource)
    return resources


def has_tar_suffix(filename):
    """Check if given filename suffix is a valid tar-file suffix.

    Parameters
    ----------
    filename : string
        Name of file on disk

    Returns
    -------
    Boolean
        True, if filename ends with '.tar', '.tar.gz', or '.tgz'
    """
    for suffix in ['.tar', '.tar.gz', '.tgz']:
        if filename.endswith(suffix):
            return True
    return False


def references_to_dict(elements):
    """Convery a list of HATEOAS reference objects into a dictionary.
    Parameters
    ----------
    elements : List
        List of key value pairs, i.e., [{rel:..., href:...}].
    Returns
    -------
    Dictionary
        Dictionary of rel:href pairs.
    """
    dictionary = {}
    for kvp in elements:
        dictionary[str(kvp['rel'])] = str(kvp['href'])
    return dictionary


def to_local_time(timestamp):
    """Convert a datatime object from UTC time to local time.

    Adopted from:
    http://stackoverflow.com/questions/4770297/python-convert-utc-datetime-string-to-local-datetime

    Parameters
    ----------
    timestamp : string
        Default string representation of timestamps expected to be in
        UTC time zone

    Returns
    -------
    datetime
        Datetime object in local time zone
    """
    utc = dt.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f')
    # Get UTC and local time zone
    from_zone = tz.gettz('UTC')
    to_zone = tz.tzlocal()

    # Tell the utc object that it is in UTC time zone
    utc = utc.replace(tzinfo=from_zone)

    # Convert time zone
    return utc.astimezone(to_zone)



"""Standard Cortical Observer - Web API client

Client to create, access and manipulate resources on Standard Cortical Observer
Web Servers.
"""

import atexit
import json
import os
import shutil
import tempfile
import uuid

import scoserv as sco
from scoserv import ExperimentHandle, ImageGroupHandle, ModelRunHandle, SubjectHandle


# ------------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------------

# Url of default SCO Web API hosted at NYU
DEFAULT_API = 'http://cds-swg1.cims.nyu.edu:5000/sco-server/api/v1'


# ------------------------------------------------------------------------------
#
# Standard Cortical Observer - Client API
#
# ------------------------------------------------------------------------------

class SCOClient(object):
    """Standard Cortical Observer Web API (SCO-API) Client to interact with
    existing SCO-API servers. The client provides access to SCO-API resources
    such as subjects (Anatomy MRI data), image groups, and experiments. It also
    allows to create new resources and to invoke remote model runs

    The client provides access to data files of SCO resources via a local cache.
    The user can specify a folder on disk for the cache to avoid downloading
    files inbetween instantiations of the SCO Client. Alternatively, an new
    temporary directory is used for every instance of the SCO client.
    """
    def __init__(self, api_url=None, data_dir=None):
        """Initialize the client. Provide a default SCO-API Url to use when the
        user calls client methods that list or create resources without giving
        a service base Url. The data directory will be used to cache downloaded
        files permanently. If none is given, a new temporary folder will be
        created. Note that resources in an existing cache folder will not be
        checked for consistency, i.e., they may be deleted on the originating
        server but still be accessible form the local cache. Use cache_clear()
        and cache_clear() and cache_tidy() to remove all cached resources or
        only those that no longer exist on their server.

        Parameters
        ----------
        server_url : string, optional
            Base URL of a SCO Web API. Use to create new resources when no
            API is specified.
        data_dir : string, optional
            Optional directory for caching files. Directory will be created if
            it does not exists. Uses a new temporary directory if no value is
            given by the user.
        """
        # Set the default API Url. This is the API that will be used to create
        # new resources if no API is specified explicitly.
        self.api_url = api_url if not api_url is None else DEFAULT_API
        # Set directory for file cache based on whether argument data_dir is
        # present or not
        if not data_dir is None:
            # Ensure that data_dir refers to a directory. Create the directory
            # if it does not exist
            if os.path.exists(data_dir):
                # Raise error if exsiting path is not a directory
                if not os.path.isdir(data_dir):
                    raise ValueError('not a directory: ' + data_dir)
            else:
                # Create directory if it does not exist
                os.makedirs(data_dir)
            self.directory = data_dir
        else:
            # Use temporary directory as cache directory. Register cleanup
            # handler to remove directory at exit
            self.directory = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, self.directory)
        # The cache index file is a tab delimited file with two columns:
        #
        # 1) resource url
        # 2) unique local cache identifier for resource
        self.db_file = os.path.join(self.directory, 'db.tsv')
        # If the data directory contains a cahce index file (i.e., directory is
        # existing cache directory from previous session) read file content.
        self.cache = {}
        if os.access(self.db_file, os.F_OK):
            with open(self.db_file, 'r') as f:
                for line in f:
                    tokens = line.strip().split('\t')
                    self.cache[tokens[0]] = tokens[1]
        # Set a local cache for SCO API links
        self.apis = {}

    def cache_add(self, resource_url, cache_id):
        """Add entry permanently to local cache.

        Parameters
        ----------
        resource_url : string
            Resource Url
        cache_id : string
            Unique cache identifier for resource
        """
        # Add entry to cache index
        self.cache[resource_url] = cache_id
        # Write cache index content to database file
        with open(self.db_file, 'w') as f:
            for resource in self.cache:
                f.write(resource + '\t' + self.cache[resource] + '\n')

    def cache_clear(self):
        """Clear local cache by deleting all cached resources and their
        downloaded files.
        """
        # Delete content of local cache directory
        for f in os.listdir(self.directory):
            f = os.path.join(self.directory, f)
            if os.path.isfile(f):
                os.remove(f)
            elif os.path.isdir(f):
                shutil.rmtree(f)
        # Empty cache index
        self.cache = {}

    def get_api_references(self, api_url=None):
        """Get set of HATEOAS reference for the given SCO-API. Use the default
        SCO-API if none is given. References are cached as they are not expected
        to change.

        Parameters
        ----------

        Returns
        -------
        """
        # Get subject listing Url for SCO-API
        if not api_url is None:
            url = api_url
        else:
            url = self.api_url
        # Check if API references are in local cache. If not send GET request
        # and add the result to the local cache
        if not url in self.apis:
            self.apis[url] = sco.references_to_dict(
                sco.JsonResource(url).json[sco.REF_LINKS]
            )
        return self.apis[url]

    # --------------------------------------------------------------------------
    # Experiments
    # --------------------------------------------------------------------------

    def experiments_create(self, name, subject_id, image_group_id, api_url=None, properties=None):
        """Create a new experiment at the given SCO-API. Subject and image
        group reference existing resources at the SCO-API.

        Parameters
        ----------
        name : string
            User-defined name for experiment
        subject_id : string
            Unique identifier for subject at given SCO-API
        image_group_id : string
            Unique identifier for image group at given SCO-API
        api_url : string, optional
            Base Url of SCO-API where experiment will be created
        properties : Dictionary, optional
            Set of additional properties for created experiment. The given
            experiment name will override an existing name property in this set.

        Returns
        -------
        scoserv.ExperimentHandle
            Handle for local copy of created experiment resource
        """
        # Create experiment and return handle for created resource
        return self.experiments_get(
            sco.ExperimentHandle.create(
                self.get_api_references(api_url)[sco.REF_EXPERIMENTS_CREATE],
                name,
                subject_id,
                image_group_id,
                properties=properties
            )
        )

    def experiments_get(self, resource_url):
        """Get handle for experiment resource at given Url.

        Parameters
        ----------
        resource_url : string
            Url for experiment resource at SCO-API

        Returns
        -------
        scoserv.ExperimentHandle
            Handle for local copy of experiment resource
        """
        # Get resource directory, Json representation, active flag, and cache id
        obj_dir, obj_json, is_active, cache_id = self.get_object(resource_url)
        # Create experiment handle. Will raise an exception if resource is not
        # in cache and cannot be downloaded.
        experiment = ExperimentHandle(obj_json, self)
        # Add resource to cache if not exists
        if not cache_id in self.cache:
            self.cache_add(resource_url, cache_id)
        # Return experiment handle
        return experiment

    def experiments_list(self, api_url=None, offset=0, limit=-1, properties=None):
        """Get list of experiment resources from a SCO-API.

        Parameters
        ----------
        api_url : string, optional
            Base Url of the SCO-API. Uses default API if argument not present.
        offset : int, optional
            Starting offset for returned list items
        limit : int, optional
            Limit the number of items in the result
        properties : List(string)
            List of additional object properties to be included for items in
            the result

        Returns
        -------
        List(scoserv.ResourceHandle)
            List of resource handles (one per image group in the listing)
        """
        # Get subject listing Url for given SCO-API and return the retrieved
        # resource listing
        return sco.get_resource_listing(
            self.get_api_references(api_url)[sco.REF_EXPERIMENTS_LISTING],
            offset,
            limit,
            properties
        )

    # --------------------------------------------------------------------------
    # Model Runs
    # --------------------------------------------------------------------------

    def experiments_runs_create(self, name, api_url, arguments={}, properties=None):
        """Create a new model run at the given SCO-API.

        Parameters
        ----------
        name : string
            User-defined name for experiment
        api_url : string
            Url to POST create model run request
        arguments : Dictionary
            Dictionary of arguments for model run
        properties : Dictionary, optional
            Set of additional properties for created mode run.

        Returns
        -------
        scoserv.ModelRunHandle
            Handle for local copy of created model run resource
        """
        # Create experiment and return handle for created resource
        return self.experiments_runs_get(
            sco.ModelRunHandle.create(
                api_url,
                name,
                arguments,
                properties=properties
            )
        )

    def experiments_runs_get(self, resource_url):
        """Get handle for model run resource at given Url.

        Parameters
        ----------
        resource_url : string
            Url for model run resource at SCO-API

        Returns
        -------
        scoserv.ModelRunHandle
            Handle for local copy of model run resource
        """
        # Get resource directory, Json representation, active flag, and cache id
        obj_dir, obj_json, is_active, cache_id = self.get_object(resource_url)
        # Create model run handle. Will raise an exception if resource is not
        # in cache and cannot be downloaded.
        run = ModelRunHandle(obj_json, obj_dir, self)
        # Add resource to cache if not exists
        if not cache_id in self.cache:
            self.cache_add(resource_url, cache_id)
        # Return model run handle
        return run

    def experiments_runs_list(self, listing_url, offset=0, limit=-1, properties=None):
        """Get list of experiment resources from a SCO-API.

        Parameters
        ----------
        listing_url : string
            url for experiments run listing.
        offset : int, optional
            Starting offset for returned list items
        limit : int, optional
            Limit the number of items in the result
        properties : List(string)
            List of additional object properties to be included for items in
            the result

        Returns
        -------
        List(scoserv.ModelRunDescriptor)
            List of model run descriptors
        """
        return sco.get_run_listing(
            listing_url,
            offset=offset,
            limit=limit,
            properties=properties
        )

    def experiments_runs_update_state_active(self, resource_url):
        """Update state of model run resource at given Url to 'ACTIVE'.

        Parameters
        ----------
        resource_url : string
            Url for model run resource at SCO-API

        Returns
        -------
        scoserv.ModelRunHandle
            Handle for local copy of model run resource
        """
        # Send state update request.
        return self.experiments_runs_get(resource_url).update_state_active()


    # --------------------------------------------------------------------------
    # General
    # --------------------------------------------------------------------------

    def get_object(self, resource_url):
        """Get remote resource information. Creates a local directory for the
        resource if this is the first access to the resource. Downloads the
        resource Json representation and writes it into a .json file in the
        cache directory.

        Raises ValueError if resource is not cached and does not exist. If the
        resource no longer exists on the server but in the local cache, a
        reference to the local copy is returned and the value of the is_active
        flag is False.

        Parameters
        ----------
        cache_id : string
            Unique cache identifier
        resource_url : string
            Url of the resource

        Returns
        -------
        (string, Json, Boolean, string)
            Returns a 4-tuple containing local resource directory, the Json
            object representing the resource, an active flag indicating if
            the resource still exists on the remote server or only in the local
            cache, and the resource unique cache identifier.
        """
        # Check if resource is in local cache. If not, create a new cache
        # identifier and set is_cached flag to false
        if resource_url in self.cache:
            cache_id = self.cache[resource_url]
        else:
            cache_id = str(uuid.uuid4())
        # The local cahce directory for resource is given by cache identifier
        obj_dir = os.path.join(self.directory, cache_id)
        # File for local copy of object's Json representation
        f_json = os.path.join(obj_dir, '.json')
        # Object active flag
        is_active = True
        # Read the remote resource representation
        try:
            obj_json = sco.JsonResource(resource_url).json
            # Save local copy of Json object. Create local resource directory if
            # it doesn't exist
            if not os.path.isdir(obj_dir):
                os.mkdir(obj_dir)
            with open(f_json, 'w') as f:
                json.dump(obj_json, f)
        except ValueError as ex:
            # If the resource does not exists but we have a local copy then read
            # object from local disk. Set is_active flag to false. Raise
            # ValueError if no local copy exists
            if os.path.isfile(f_json):
                with open(f_json, 'r') as f:
                    obj_json = json.load(f)
                is_active = False
            else:
                raise ex
        # Return object directory, Json, active flag, and cache identifier
        return obj_dir, obj_json, is_active, cache_id

    # --------------------------------------------------------------------------
    # Image Groups
    # --------------------------------------------------------------------------

    def image_groups_create(self, filename, api_url=None, options=None, properties=None):
        """Create new image group at given SCO-API by uploading local file.
        Expects an tar-archive containing images in the image group. Allows to
        update properties of created resource.

        Parameters
        ----------
        filename : string
            Path to tar-archive on local disk
        api_url : string, optional
            Base Url of SCO-API where image group will be created
        options : Dictionary, optional
            Values for image group options
        properties : Dictionary, optional
            Set of additional properties for created image group

        Returns
        -------
        scoserv.ImageGroupHandle
            Handle for local copy of created image group resource
        """
        # Create image group and return handle for created resource
        return self.image_groups_get(
            sco.ImageGroupHandle.create(
                self.get_api_references(api_url)[sco.REF_IMAGE_GROUPS_CREATE],
                filename,
                options,
                properties
            )
        )

    def image_groups_get(self, resource_url):
        """Get handle for image group resource at given Url.

        Parameters
        ----------
        resource_url : string
            Url for image group resource at SCO-API

        Returns
        -------
        scoserv.ImageGroupHandle
            Handle for local copy of image group resource
        """
        # Get resource directory, Json representation, active flag, and cache id
        obj_dir, obj_json, is_active, cache_id = self.get_object(resource_url)
        # Create image group handle. Will raise an exception if resource is not
        # in cache and cannot be downloaded.
        image_group = ImageGroupHandle(obj_json, obj_dir)
        # Add resource to cache if not exists
        if not cache_id in self.cache:
            self.cache_add(resource_url, cache_id)
        # Return image group handle
        return image_group

    def image_groups_list(self, api_url=None, offset=0, limit=-1, properties=None):
        """Get list of image group resources from a SCO-API.

        Parameters
        ----------
        api_url : string, optional
            Base Url of the SCO-API. Uses default API if argument not present.
        offset : int, optional
            Starting offset for returned list items
        limit : int, optional
            Limit the number of items in the result
        properties : List(string)
            List of additional object properties to be included for items in
            the result

        Returns
        -------
        List(scoserv.ResourceHandle)
            List of resource handles (one per image group in the listing)
        """
        # Get subject listing Url for given SCO-API and return the retrieved
        # resource listing
        return sco.get_resource_listing(
            self.get_api_references(api_url)[sco.REF_IMAGE_GROUPS_LIST],
            offset,
            limit,
            properties
        )

    # --------------------------------------------------------------------------
    # Subjects
    # --------------------------------------------------------------------------

    def subjects_create(self, filename, api_url=None, properties=None):
        """Create new anatomy subject at given SCO-API by uploading local file.
        Expects an tar-archive containing a FreeSurfer anatomy.

        Parameters
        ----------
        filename : string
            Path to tar-archive on local disk
        api_url : string, optional
            Base Url of SCO-API where subject will be created
        properties : Dictionary, optional
            Set of additional properties for created subject

        Returns
        -------
        scoserv.SubjectHandle
            Handle for local copy of created image group resource
        """
        # Create image group and return handle for created resource
        return self.subjects_get(
            sco.SubjectHandle.create(
                self.get_api_references(api_url)[sco.REF_SUBJECTS_CREATE],
                filename,
                properties
            )
        )

    def subjects_get(self, resource_url):
        """Get handle for subject resource at given Url.

        Parameters
        ----------
        resource_url : string
            Url for subject resource at SCO-API

        Returns
        -------
        scoserv.SubjectHandle
            Handle for local copy of subject resource
        """
        # Get resource directory, Json representation, active flag, and cache id
        obj_dir, obj_json, is_active, cache_id = self.get_object(resource_url)
        # Create subject handle. Will raise an exception if resource is not
        # in cache and cannot be downloaded.
        subject = SubjectHandle(obj_json, obj_dir)
        # Add resource to cache if not exists
        if not cache_id in self.cache:
            self.cache_add(resource_url, cache_id)
        # Return subject handle
        return subject

    def subjects_list(self, api_url=None, offset=0, limit=-1, properties=None):
        """Get list of subject resources from a SCO-API.

        Parameters
        ----------
        api_url : string, optional
            Base Url of the SCO-API. Uses default API if argument not present.
        offset : int, optional
            Starting offset for returned list items
        limit : int, optional
            Limit the number of items in the result
        properties : List(string)
            List of additional object properties to be included for items in
            the result

        Returns
        -------
        List(scoserv.ResourceHandle)
            List of resource handles (one per subject in the listing)
        """
        # Get subject listing Url for given SCO-API and return the retrieved
        # resource listing
        return sco.get_resource_listing(
            self.get_api_references(api_url)[sco.REF_SUBJECTS_LIST],
            offset,
            limit,
            properties
        )
